/*
 * Copyright 2016 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package criterion

import java.lang.System.nanoTime

data class Duration(val ns: Double) {
    val ms: Double
        get() = ns / 1e+6
}

inline
fun durationOf(block: () -> Unit): Duration =
    nanoTimeOf(block).let {
        Duration(it.toDouble())
    }

inline
fun nanoTimeOf(block: () -> Unit): Long =
    nanoTime().let { start ->
        block()
        nanoTime() - start
    }
